<?php

namespace App\Http\Controllers;

use App\Http\Requests\CreateApprovalRequest;
use App\Http\Requests\UpdateApprovalRequest;
use App\Http\Resources\ApprovalCollection;
use App\Http\Resources\ApprovalResource;
use App\Models\Approval;
use App\Services\ApprovalService;
use Illuminate\Database\Eloquent\Builder as EloquentBuilder;
use Illuminate\Database\Eloquent\Relations\Relation as EloquentRelation;
use Illuminate\Database\Query\Builder as QueryBuilder;
use Illuminate\Http\RedirectResponse;
use Illuminate\Http\Request;
use Illuminate\Http\Resources\Json\AnonymousResourceCollection;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\DB;
use Illuminate\View\View;
use Psy\Util\Json;

class ApprovalController extends Controller
{
    public function __construct(
        protected ApprovalService $approvalService)
    {
        //$this->middleware('auth');
    }

    /**
     * Apply query scopes.
     */
    protected function applyScopes(
        EloquentBuilder|QueryBuilder|EloquentRelation|Collection|AnonymousResourceCollection $query
    ): EloquentBuilder|QueryBuilder|EloquentRelation|Collection|AnonymousResourceCollection {
        foreach ($this->scopes as $scope) {
            $scope->apply($query);
        }

        return $query;
    }

    /**
     * Get the query object to be processed by datatables.
     *
     */
    public function query(): EloquentBuilder
    {
        return Approval::query();
    }
    /**
     * Display a listing of the Approval.
     *
     */

    public function index()
    {
        $approvals = Approval::select('approvals.*')->get();
        return new ApprovalCollection($approvals);
    }

    /**
     * Filter a listing of the Approval.
     *filterId ->1, needs approve by me,
     * filterId ->2, already approved by me
     * filterId ->0, all approvals
     * @param int|null $filterId
     * @return View
     */
    public function filter(int $filterId = null): View
    {
        $this->approvalService->setApprovalFilter($filterId);
        $this->approvalService->limit = 20;

        if($filterId == 2) {
            $this->approvalService->joins = ['approval_histories' => ['approval_histories.approval_id','=','approvals.id']];
        }

        $approvals = $this->approvalService->findAll();

        return view('approvals.index', compact('approvals'));
    }

    /**
     * @param Request $request
     */

    public function approve(Request $request)
    {
        /*
         * Expected inputs from user
         {
            "_token": "EYNR1jeBnmGudEzDOKsXeDBvyLRheT1udcoTiKvs",
            "action": "approved",
            "approved_by": 1,
            "approval_id": 1,
            "comment": "er"
         }
        */
        $input = $request->all();
        $this->approvalService->setApprovalId($input['approval_id']);
        $this->approvalService->setInputs($input);
        $approval = $this->approvalService->approve();
        if($approval instanceof Approval) {
            return new ApprovalResource($approval);
        }
        else{
            return $approval;
        }



    }

    /**
     * Show the form for creating a new Approval.
     *
     * @return View
     */
    public function create(): View
    {
        return view('approvals.create');
    }

    /**
     * Store a newly created Approval in storage.
     *
     * @param CreateApprovalRequest $request
     *
     */
    public function store(CreateApprovalRequest $request)
    {
        $input = $request->all();
//        return response()->json([
//            "data" => $input
//        ]);

        $approval = $this->approvalService->createApproval($input);

        return new ApprovalResource($approval);
    }

    /**
     * Display the specified Approval.
     *
     * @param int $id
     *
     */
    public function show(int $id)
    {
        $approval = $this->approvalService->getApprovalById($id);

        if (empty($approval)) {
            //redirect to approvals which need my approval
            return redirect(route('approvals.filter', 1));
        }

       // $approvable = $approval->approvable;
        $approvalHistories = $approval->approvalHistories;

        return new ApprovalResource($approval);
    }

    /**
     * Show the form for editing the specified Approval.
     *
     * @param int $id
     *
     * @return RedirectResponse | View
     */
    public function edit(int $id): View|RedirectResponse
    {
        $approval = $this->approvalService->getApprovalById($id);

        if (empty($approval)) {
            Alert::error('Not found','Approval not found');

            return redirect(route('approvals.index'));
        }

        return view('approvals.edit')->with('approval', $approval);
    }

    /**
     * Update the specified Approval in storage.
     *
     * @param int $id

     */
    public function update(int $id, UpdateApprovalRequest $request)
    {
        $approval = $this->approvalService->getApprovalById($id);

        if (empty($approval)) {
            return json_encode(['status' => 'error', 'message' => 'Approval not found','code' => 404]);
        }
        $this->approvalRepository->update($request->all(), $id);

        return json_encode(['status' => 'success', 'message' => 'Approval updated successfully.','code' => 200]);
    }

    public function getApprovalWorkFlowByApprovalId(int $approvalId){
        $approval = Approval::with('approvalWorkFlows')->where("id",$approvalId)->first();
        $approvalRoles = $approval->getRolesRequiredToApprove();
        $approvedRoles = $approval->approvedRoles();

        return json_encode([
            'approval' => $approval,
            'approval_roles' => $approvalRoles,
            'approved_roles' => $approvedRoles,
        ]);
    }

    /**
     * Remove the specified Approval from storage.
     */
    public function destroy(int $id)
    {
        $approval = $this->approvalService->getApprovalById($id);

        if (empty($approval)) {
            return json_encode(['status' => 'error', 'message' => 'Approval not found','code' => 404]);
        }

        $this->approvalService->deleteAApproval($id);

        return json_encode(['status' => 'success', 'message' => 'Approval deleted successfully.','code' => 200]);
    }
}
