<?php

namespace App\Repositories;

use Exception;
use Illuminate\Container\Container as Application;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Model;


abstract class BaseRepository
{
    /**
     * @var Model
     */
    protected $model;

    /**
     * @var Application
     */
    protected $app;

    /**
     * @param Application $app
     *
     * @throws Exception
     */
    public function __construct(Application $app)
    {
        $this->app = $app;
        $this->makeModel();
    }

    /**
     * Get searchable fields array
     *
     * @return array
     */
    abstract public function getFieldsSearchable();

    /**
     * Get joinable fields array
     *
     * @return array
     */
    abstract public function getTablesJoinable();

    /**
     * Configure the Model
     *
     * @return string
     */
    abstract public function model();

    /**
     * Make Model instance
     *
     * @throws Exception
     *
     * @return Model
     */
    public function makeModel()
    {
        $model = $this->app->make($this->model());

        if (!$model instanceof Model) {
            throw new Exception("Class {$this->model()} must be an instance of Illuminate\\Database\\Eloquent\\Model");
        }

        return $this->model = $model;
    }

    /**
     * Paginate records for scaffold.
     *
     * @param int $perPage
     * @param array $columns
     * @return \Illuminate\Contracts\Pagination\LengthAwarePaginator
     */
    public function paginate($perPage, $columns = ['*'], $order = [])
    {
        //dd($order);
        $query = $this->allQuery(order: $order);

        return $query->paginate($perPage, $columns);
    }

    /**
     * Build a query for retrieving all records.
     *
     * @param array|null $search
     * @param int|null $skip
     * @param int|null $limit
     * @param array|null $joins
     * @return Builder
     */
    public function allQuery(array $search = null, int $skip = null, int $limit = null, array $joins = null, $order = null): Builder
    {
        $query = $this->model->newQuery();

        if (!is_null($search))
        {
            foreach($search as $key => $value) {
                if (in_array($key, $this->getFieldsSearchable())) {
                    $query->where($key, $value);
                }
            }
        }

        if(!is_null($joins))
        {
            foreach ($joins as $key => $value){
                if(in_array($key,$this->getTablesJoinable())){
                    $query->join($key,$value[0],$value[1],$value[2]);
                }
            }
        }

        if (!is_null($skip)) {
            $query->skip($skip);
        }

        if (!is_null($limit)) {
            $query->limit($limit);
        }

        if($order){
            $query->orderBy($order[0],$order[1]);
        }

        return $query;
    }

    /**
     * Retrieve all records with given filter criteria
     *
     * @param array $search
     * @param int|null $skip
     * @param int|null $limit
     * @param array $columns
     *
     */
    public function all($search = [], $skip = null, $limit = null, $columns = ['*'],$joins = [])
    {
        $query = $this->allQuery($search, $skip, $limit, $joins);
        return $query->get($columns);
    }

    /**
     * instantiate new model instance with or without inputs
     * @param array $input user inputs
     * @return Model
     */
    public function newModelInstance(array $input = []): Model
    {
        return $this->model->newInstance($input);
    }

    /**
     * Create model record
     *
     * @param array $input
     *
     * @return Model
     */
    public function create($input)
    {
        $model = $this->model->newInstance($input);

        $model->save();

        return $model;
    }

    /**
     * Create new model record with related model
     *
     * @param array $input user inputs
     * @param Model $parentModel parent related model
     * @param string $relationship child relationship name from relate model instance
     * @return Model new model instance tobe created
     */
    public function createWithRelation(array $input, Model $parentModel, string $relationship): Model
    {
        $model = $this->model->newInstance($input);

        return $parentModel->{$relationship}()->save($model);
    }

    /**
     * Find model record for given id
     *
     * @param int $id
     * @param array $columns
     *
     * @return Builder|Builder[]|\Illuminate\Database\Eloquent\Collection|Model|null
     */
    public function find($id, $columns = ['*'])
    {
        $query = $this->model->newQuery();

        return $query->find($id, $columns);
    }

    /**
     * Update model record for given id
     *
     * @param array $input
     * @param int $id
     *
     * @return Builder|Builder[]|\Illuminate\Database\Eloquent\Collection|Model
     */
    public function update($input, $id)
    {
        $query = $this->model->newQuery();

        $model = $query->findOrFail($id);

        $model->fill($input);

        $model->save();

        return $model;
    }

    /**
     * @param int $id
     *
     * @throws Exception
     *
     * @return bool|mixed|null
     */
    public function delete($id)
    {
        $query = $this->model->newQuery();

        $model = $query->findOrFail($id);

        return $model->delete();
    }
}
