<?php

namespace App\Services;

use App\Enums\ApprovalStatus;
use App\Events\ApprovalActionCreated;
use App\Models\Approval;
use App\Models\ApprovalHistory;
use App\Models\ApprovalWorkFlow;
use App\Repositories\ApprovalRepository;
use App\Traits\FilterPropertiesTrait;
use Exception;
use Illuminate\Database\Eloquent\ModelNotFoundException;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Log;

class ApprovalService
{
    use FilterPropertiesTrait;

    private int $filterId, $approvalId;
    private array $input=[];

    /**
     * @param int $filterId
     */
    public function __construct(protected ApprovalRepository $approvalRepository){}

    public function findAll(){
        $approvals = $this->approvalRepository->all($this->getSearchFilter(),null, $this->limit,['*'],$this->joins);
        //$approvals = Approval::select('approvals.*');

        if ($this->getApprovalFilter() == 1) {
            $approvals->where('is_approved', '=', 0)
                ->whereIn('current_approval_role_id', Auth::user()->roles->pluck('id'))
                ->where('status', '!=', ApprovalStatus::STATUS_REJECTED->value);
            //dd(ApprovalStatus::STATUS_REJECTED->value);
        }
//        //already approved by me
        if ($this->getApprovalFilter() == 2) {
            $approvals
                ->where('approvals.is_approved', '=', true)
                ->whereIn('approval_histories.role_id', Auth::user()->roles->pluck('id'));
        }
       // dd($approvals);
        return $approvals;
    }

    public function approve(){
        $approval = $this->approvalRepository->find($this->approvalId);

        if (empty($approval)) {
            $response = ['message' => 'Approval not found', 'code' => 404];
            return json_encode($response);
        }
        //begin transaction
        DB::beginTransaction();
        try {
            if(!$approval->is_approved){
                ApprovalHistory::create([
                    'is_approved' => $this->input['action'] == 'approved',
                    'comment' => $this->input['comment'],
                    'approved_by' => $this->input['approved_by'],
                    'approval_id' => $approval->id,
                    'role_id' => $approval->current_approval_role_id,
                ]);
            }

            //refresh approval
            $approval = $this->approvalRepository->find($approval->id);
            if (!$approval->isApproved() && !$approval->isRejected()) {
                $approval->update(['status' => ApprovalStatus::STATUS_PROGRESSING->value]);
                if($this->input['action'] == 'approved'){
                    $approval->updateCurrentApprovers();
                }
            }

            elseif ($approval->isApproved() && !$approval->isRejected()) {
                $approval->update(['is_approved' => true, 'status' => ApprovalStatus::STATUS_APPROVED->value]);
            }

            elseif ($approval->isRejected()) {
                $approval->update(['status' => ApprovalStatus::STATUS_REJECTED->value]);
            }
        }
        catch (ModelNotFoundException $e) {
            DB::rollback();
            Log::info("fail to approval history: " . print_r(get_class_methods($e), true));
        }
        DB::commit();
        return $approval;
    }

    public function setApprovalFilter($filterId): void
    {
        $this->filterId = $filterId;
    }
    public function getApprovalFilter(): int
    {
        return $this->filterId;
    }
    public function setApprovalId($approvalId): void
    {
        $this->approvalId = $approvalId;
    }
    public function getApprovalId(): int
    {
        return $this->approvalId;
    }
    public function setInputs($inputs): void
    {
        $this->input = $inputs;
    }
    public function getInputs(): array
    {
        return $this->input;
    }

    public function getApprovalById(int $approvalId): Approval
    {
        return $this->approvalRepository->find($approvalId);
    }

    public function createApproval($input){
        $approval = $this->approvalRepository->create($input);
        $approval->update(['status' => ApprovalStatus::STATUS_WAITING_FOR_APPROVAL->value]);
        $workFlow = ApprovalWorkFlow::find($input['current_approval_work_flow_id']);
        $approval->approvalWorkFlows()->attach($workFlow);
        return $approval;
    }

    /**
     * @throws Exception
     */
    public function deleteAApproval($id): void
    {
        $approval = $this->approvalRepository->find($id);
        $approval->approvalWorkFlows()->detach();
        $approval->approvalHistories()->delete();
        $this->approvalRepository->delete($id);
    }
}
